local Bebo = {}
Bebo.__index = Bebo
setmetatable(Bebo, Mob)

function Bebo:create (x, y, z, map)
	local bebo = {}
	setmetatable(bebo, Bebo)
	bebo.model = Snek_3d:create(MODELS.dog)
	bebo.map = map
	bebo.pos = {
		x = x,
		y = y,
		z = z,
	}
	bebo.vel = {
		x = 0,
		y = 0,
		z = 0,
	}
	bebo.jump_force = 10 --tile height units per second
	bebo.on_floor = false
	bebo.gravity_force = 35 --tile height units per second
	bebo.deceleration = .0002
	bebo.acc = 50
	bebo.width = .75
	bebo.height = .75
	bebo.model.rotation_x = -math.pi * .2
	return bebo
end

function Bebo:set_camera ()
	local x, y = self.pos.x, self.pos.y
	local size = self.map.tilesize
	local dx = (x / 2 - y / 2) * size
	local dy = (y / 4 + x / 4) * size
	self.map.offset_x = -dx + LG.getWidth() / 2
	self.map.offset_y = -dy + LG.getHeight() / 2
end

function Bebo:update (dt)
	self:set_camera()
	self.model:update()
	local acc = self.acc * dt
	local vx, vy = self.vel.x, self.vel.y

	if INPUT:down("left") then
		vx = vx - acc
		vy = vy + acc
	elseif INPUT:down("right")then
		vx = vx + acc
		vy = vy - acc
	end
	if INPUT:down("up") then
		vy = vy - acc
		vx = vx - acc
	elseif INPUT:down("down")then
		vy = vy + acc
		vx = vx + acc
	end

	local dir = UTIL.get_rotation(vy, vx)
	self.model.rotation_y = dir + math.pi * -.25

	if INPUT:down("jump")then
		self:jump(dt)
	end

	self.vel.x = vx
	self.vel.y = vy
	self.pos.x = self.pos.x + self.vel.x * dt
	self.pos.y = self.pos.y + self.vel.y * dt
	self.pos.z = self.pos.z + self.vel.z * dt

	self:momentum(dt)
	self:gravity(dt)
	self:bounds_collision(dt)
	self:collision_resolution(dt)
end

function Bebo:jump (dt)
	local tx, ty = math.floor(self.pos.x), math.floor(self.pos.y)
	if self.on_floor then
		self.vel.z = self.jump_force
	end
end

function Bebo:momentum (dt)
	local tolerance = .001
	self.vel.x = self.vel.x * math.pow(self.deceleration, dt);
	self.vel.y = self.vel.y * math.pow(self.deceleration, dt);
	if math.abs(self.vel.x) < tolerance then
		self.vel.x = 0
	end
	if math.abs(self.vel.y) < tolerance then
		self.vel.y = 0
	end
	if math.abs(self.vel.z) < tolerance then
		self.vel.z = 0
	end
end

function Bebo:gravity (dt)
	local x, y = self.pos.x, self.pos.y
	local w, h = self.width / 2, self.height / 2
	local min_x, min_y = math.floor(x - w), math.floor(y - h)
	local max_x, max_y = math.floor(x + w), math.floor(y + h)
	local on_floor = false
	local elevation = 0
	local tolerance = 5 * dt

	for ty = min_y, max_y do
		for tx = min_x, max_x do
			elevation = self.map:get_elevation(tx, ty)
			if math.abs(self.pos.z - elevation) <= tolerance then
				on_floor = true
				break
			end
			elevation = self.pos.z
		end
		if on_floor then
			break
		end
	end

	if on_floor and self.vel.z < 0 then
		self.on_floor = true
		self.pos.z = math.floor(elevation)
		self.vel.z = 0
	else
		self.on_floor = false
		self.vel.z = self.vel.z - self.gravity_force * dt
	end
end

function Bebo:bounds_collision (dt)
	local x, y = self.pos.x, self.pos.y
	local w, h = self.width / 2, self.height / 2
	local map_w, map_h = self.map.width, self.map.height
	if x < w then
		self.pos.x = w
	elseif x > map_w - w then
		self.pos.x = map_w - w
	end
	if y < h then
		self.pos.y = h
	elseif y > map_h - h then
		self.pos.y = map_h - h
	end
end

function Bebo:collision_resolution (dt)
	local x, y = self.pos.x, self.pos.y
	local w, h = self.width / 2, self.height / 2
	local min_x, min_y = math.floor(x - w), math.floor(y - h)
	local max_x, max_y = math.floor(x + w), math.floor(y + h)
	for ty = min_y, max_y do
		for tx = min_x, max_x do
			if self.map:get_elevation(tx, ty) > self.pos.z then
				local cx = tx + .5
				local cy = ty + .5
				local pen_x = 0
				local pen_y = 0
				if x > cx then
					pen_x = (tx + 1) - (x - w)
				else
					pen_x = tx - (x + w)
				end

				if y > cy then
					pen_y = (ty + 1) - (y - h)
				else
					pen_y = ty - (y + h)
				end

				if math.abs(pen_x) < math.abs(pen_y) then
					if x > cx then
						self.pos.x = tx + 1 + w
					else
						self.pos.x = tx - w
					end
				else
					if y > cy then
						self.pos.y = ty + 1 + h
					else
						self.pos.y = ty - h
					end
				end
			end
		end
	end
end

function Bebo:draw_collision_hints ()
	LG.push()
	LG.reset()
	local x, y = self.map:tile_pos_to_screen(
		self.pos.x,
		self.pos.y
	)
	LG.circle("fill", x, y, 3)
	local w, h = self.width * self.map.tilesize / 2, self.height * self.map.tilesize / 2
	LG.translate(x, y)
	LG.translate(0, -w / 2)
	LG.scale(1, .5)
	LG.rotate(math.pi / 4)
	LG.rectangle(
		"line",
		0,
		0,
		w * 1.5,
		h * 1.5
	)
	LG.pop()
end

function Bebo:draw ()
	LG.push()
	LG.scale(.3)
	--self:draw_collision_hints()
	self.model:draw()
	LG.pop()
end

return Bebo
