local LG = love.graphics
local Dimetric_Map = {}
Dimetric_Map.__index = Dimetric_Map

function Dimetric_Map:create(level)
	self:check_level_integrity(level)
	local map = {
		tilesize = level.tilesize,
		block_height = level.block_height,
		width = level.width or 1,
		height = level.height or 1,
		offset_x = 0,
		offset_y = 0,
		tile_map = level.tile_map,
		sprites = Dimetric_Map:get_sprites_from_level(level),
		spritesheet = level.spritesheet,
		max_elevation = 0,
		max_elevation_surplus = level.max_elevation_surplus or 0,
		objects = {},
		object_canvas = LG.newCanvas(
			level.tilesize,
			level.tilesize
		)
	}
	setmetatable(map, Dimetric_Map)
	map:set_max_elevation()
	return map
end

function Dimetric_Map:check_level_integrity(level)
	local error_prefix = "Dimetric Map Error: "
	local sw, sh = level.spritesheet:getWidth(), level.spritesheet:getHeight()
	local dimension_suggestion = "Check if you have any extra padding on the spritesheet, if anything is cut off,\nor if your proportions are off."
	if (not level) then
		error(error_prefix .. "No level provided")
	end
	if (not level.block_height) then
		error(error_prefix .. "Level block_height not set")
	end
	if (not level.tilesize) then
		error(error_prefix .. "Level tilesize not set")
	end
	if (level.block_height < level.tilesize / 2) then
		error(error_prefix .. "Block height can't be smaller than half the tilesize.")
	end
	if not level.tiles_x and (sw % level.tilesize ~= 0) then
		error(error_prefix .. "Spritesheet width not evenly divisible by tilesize.\n" .. dimension_suggestion)
	end
	if not level.tiles_y and (sh % level.block_height ~= 0) then
		error(error_prefix .. "Spritesheet height not evenly divisible by tilesize.\n" .. dimension_suggestion)
	end

	if (#level.tile_map ~= level.width * level.height) then
		error(error_prefix .. "Level tile_map does not match width and height")
	end
end

function Dimetric_Map:get_sprites_from_level(level)
	local sprites = {}
	local spritesheet = level.spritesheet
	local spritesheet_width, spritesheet_height  = spritesheet:getWidth(), spritesheet:getHeight()
	local block_width = level.tilesize
	local block_height = level.block_height
	local tiles_x = level.tiles_x or math.floor(spritesheet:getWidth() / block_width)
	local tiles_y = level.tiles_y or math.floor(spritesheet:getHeight() / block_height)

	spritesheet:setFilter("nearest", "nearest")

	for y = 0, tiles_y - 1 do
		local offset_y = y * block_height
		for x = 0, tiles_x - 1 do
			local offset_x = x * block_width
			local quad = LG.newQuad(
				offset_x,
				offset_y,
				block_width,
				block_height,
				spritesheet_width,
				spritesheet_height
			)
			sprites[#sprites + 1] = quad
		end
	end
	return sprites
end

function Dimetric_Map:set_max_elevation()
	local max_elevation = 0
	for x = 0, self.width do
		for y = 0, self.height do
			local elevation = self:get_elevation(x, y)
			if elevation > max_elevation then
				max_elevation = elevation
			end
		end
	end
	self.max_elevation = max_elevation + self.max_elevation_surplus
end

function Dimetric_Map:add_object(pos, width, height, draw)
	self.objects[#self.objects + 1] = {
		pos = pos,
		width = width,
		height = height,
		draw = draw,
	}
end

function Dimetric_Map:get_obstructing_tiles(x, y, z)
	local tiles = {}
	local max_elevation = math.max(math.ceil(z) + 1, self.max_elevation + self.max_elevation_surplus)
	local fz = math.floor(z)
	local cz = math.floor(z)
	for i = 1, max_elevation do
		x = x + 1
		y = y + 1
		local elevation = self:get_elevation(x, y)
		if elevation >= z then
			for j = 0, (elevation - fz) - 1 do
				tiles[#tiles + 1] = {x, y, fz + j}
			end
		end
		elevation = self:get_elevation(x, y - 1)
		if elevation >= z then
			for j = 0, (elevation - fz) - 1 do
				tiles[#tiles + 1] = {x, y - 1, fz + j}
			end
		end
		elevation = self:get_elevation(x - 1, y)
		if elevation >= z then
			for j = 0, (elevation - fz) - 1 do
				tiles[#tiles + 1] = {x - 1, y, fz + j}
			end
		end
	end
	return tiles
end

function Dimetric_Map:draw_object(object)
	local tx, ty, tz = object.pos.x, object.pos.y, object.pos.z
	local x, y = self:tile_pos_to_screen(tx, ty)
	local block_base_height = self:get_block_base_height()
	local c = self.object_canvas
	local cw, ch = c:getWidth(), c:getHeight()
	local obstructing_tiles = self:get_obstructing_tiles(tx, ty, tz)
	local original_offset_x, original_offset_y = self.offset_x, self.offset_y
	y = y - object.pos.z * block_base_height


	LG.push()
	LG.setCanvas(c)
	LG.clear(0, 0, 0, 0, true, true)
	LG.translate(cw / 2, ch / 2)
	for i = 1, #obstructing_tiles do
		local ttx, tty, ttz = math.floor(obstructing_tiles[i][1]), math.floor(obstructing_tiles[i][2]), math.floor(obstructing_tiles[i][3])
		local sprite_type = 5
		local tile_offset_x = ttx - tx - 1
		local tile_offset_y = tty - ty - 1
		local tile_offset_z = ttz - tz
		local fixed_offset_y = self.tilesize / 8 + 2

		self.offset_x = 0
		self.offset_y = 0
		self:draw_tile(
			tile_offset_x,
			tile_offset_y,
			tile_offset_z,
			0,
			fixed_offset_y,
			sprite_type
		)
		self.offset_x, self.offset_y = original_offset_x, original_offset_y
	end

	LG.setCanvas()
	LG.pop()
	--LG.draw(c, x - cw / 2, y - ch / 2) --honestly handy for debugging

	local mask_shader = LG.newShader([[
		extern Image mask;
		extern float mask_size;
		extern vec2 canvas_origin;

		vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords){
			vec2 mask_coords = screen_coords - canvas_origin;
			vec2 mask_uv = mask_coords / mask_size;
			vec4 mask_pixel = Texel(mask, mask_uv);
			return vec4(color.rgb, color.a - mask_pixel.a);
		}
	]])
	mask_shader:send( 'mask', c )
	mask_shader:send( 'mask_size', cw )
	mask_shader:send( 'canvas_origin', {x - cw / 2, y - cw / 2} )

	LG.translate(x, y)
	LG.setShader(mask_shader)
	object:draw()
	LG.origin()
	LG.setShader()

end

function Dimetric_Map:get_screen_width()
	return self.tilesize * self.width
end

function Dimetric_Map:get_screen_height()
	return self.tilesize * self.height / 2
end

function Dimetric_Map:get_sprite_index(x, y)
	local x = math.floor(x + 1)
	return math.floor(y) * self.width + x
end

function Dimetric_Map:get_elevation(x, y)
	local tile = self.tile_map[self:get_sprite_index(x, y)]
	if type(tile) ~= "table" then
		return 0
	end
	return #tile - 1
end

function Dimetric_Map:get_sprite_type(x, y, z)
	local tile = self.tile_map[self:get_sprite_index(x, y)]
	if type(tile) == "table" then
		return tile[z + 1] or 0
	end
	if z > 0 then
		return 0
	end
	return tile
end

function Dimetric_Map:is_tile_in_view(width, height, tile_x, tile_y, elevation)
	local x, y = self:tile_pos_to_screen(tile_x, tile_y)
	local tilesize = self.tilesize
	return
		x > -self.tilesize / 2 and
		y > -self.tilesize / 2 and
		x < width + self.tilesize / 2 and
		y < height + self.tilesize / 2 + elevation * self.block_height
end

function Dimetric_Map:screen_to_tile_pos(x, y)
	local ox, oy = self.offset_x, self.offset_y
	x = x - ox
	y = y - oy
	local size = self.tilesize
	local dx = (y * 2 + x) / size
	local dy = (y * 2 - x) / size
	return dx, dy
end

function Dimetric_Map:get_block_base_height()
	return self.block_height - self.tilesize / 2
end

function Dimetric_Map:tile_pos_to_screen(x, y)
	local ox, oy = self.offset_x, self.offset_y
	local size = self.tilesize
	local dx = ((x / 2 - y / 2) * size)
	local dy = ((y / 4 + x / 4) * size)
	dx = dx + ox
	dy = dy + oy
	return dx, dy
end

function Dimetric_Map:draw_tile(tile_x, tile_y, tile_z, draw_offset_x, draw_offset_y, sprite_type)
	if sprite_type < 1 then
		return
	end
	local x, y = self:tile_pos_to_screen(tile_x, tile_y)
	local ox = draw_offset_x or 0
	local oy = draw_offset_y or 0
	local size = self.tilesize
	local block_height = self.block_height - self.tilesize / 2
	local quad = self.sprites[sprite_type]
	x = x + ox
	y = y + oy
	x = x - size / 2
	y = y - block_height * tile_z

	LG.setColor(1, 1, 1)
	LG.draw(
		self.spritesheet,
		quad,
		math.floor(x),
		math.floor(y)
	)
end

function Dimetric_Map:get_objects_on_tile(x, y, z) local objects = {}
	for i = 1, #self.objects do
		local object = self.objects[i]
		if
			object.pos.z >= z and
			object.pos.x >= x and object.pos.x < x + 1 and
			object.pos.y >= y and object.pos.y < y + 1
		then
			objects[#objects + 1] = object
		end
	end
	if #objects > 0 then
		return objects
	end
	return nil
end

function Dimetric_Map:draw(offset_x, offset_y, width, height)
	LG.setColor(1, 1, 1)
	LG.setScissor(offset_x, offset_y, width, height)
	local tiles_to_draw = {}
	local tz = 0
	local mx, my = 0, 0
	local objects = {}

	for tx = 0, self.width - 1 do
		for ty = 0, self.height - 1 do
			local elevation = self:get_elevation(tx, ty)
			if self:is_tile_in_view(width, height, tx, ty, elevation) then
				local sprite_type = self:get_sprite_type(tx, ty, tz)
				local new_objects = self:get_objects_on_tile(tx, ty, tz)
				if new_objects then
					for i = 1, #new_objects do
						objects[#objects + 1] = new_objects[i]
					end
				end
				self:draw_tile(tx, ty, tz, offset_x, offset_y, sprite_type)
				if elevation > 0 then
					mx = math.max(tx, mx)
					my = math.max(ty, my)
					tiles_to_draw[#tiles_to_draw + 1] = {tx, ty, elevation}
				end
			end
		end
	end

	for i = 1, #tiles_to_draw do
		local tx, ty, elevation = tiles_to_draw[i][1], tiles_to_draw[i][2], tiles_to_draw[i][3]
		for tz = 1, elevation do
			local sprite_type = self:get_sprite_type(tx, ty, tz)
			self:draw_tile(tx, ty, tz, offset_x, offset_y, sprite_type)
			local objects = self:get_objects_on_tile(tx, ty, tz)
		end
	end

	if objects then
		for i = 1, #objects do
			self:draw_object(objects[i])
		end
	end

	LG.setScissor()
end

return Dimetric_Map
